LISTING 3

#ifndef INCLUDED_AR_UTILITY_ABCPTR_H
#define INCLUDED_AR_UTILITY_ABCPTR_H

// copyright 1998 by anton rommerskirchen
// this code can be used and copied for free without changes
// no warrenty is given by the author
// enhancements to arommerskirchen@csi.com
// $Id: abcptr.h,v 1.1 2001/10/03 12:45:19 ar Exp $

#ifndef INCLUDED_STDEXCEPT
#	include <stdexcept>
#	undef INCLUDED_STDEXCEPT
#	define INCLUDED_STDEXCEPT
#endif

#ifndef INCLUDED_STRING
#	include <string>
#	undef INCLUDED_STRING
#	define INCLUDED_STRING
#endif

#ifndef INCLUDED_AR_UTILITY_ABCPTRMANAGER_H
#	include <ar/utility/abcptrmanager.h>
#	undef INCLUDED_AR_UTILITY_ABCPTRMANAGER_H
#	define INCLUDED_AR_UTILITY_ABCPTRMANAGER_H
#endif

#ifndef INCLUDED_AR_UTILITY_CHUNK_H
#	include <ar/utility/chunk.h>
#	undef INCLUDED_AR_UTILITY_CHUNK_H
#	define INCLUDED_AR_UTILITY_CHUNK_H
#endif

namespace AR {

namespace Utility {


//------------------------------------------------------------------------

template < class AB >
class ABCPtr
{
	// abstract base reference counted pointer -- interface seen only
	// needs help of factory classes for instantiation
	// knows only the interface class AB. the implementation is
	// not known at the interface side of the system

	Chunk* _chunk;

protected:

	explicit ABCPtr
		(
		void* i_p
		) throw ()
		:
		_chunk (reinterpret_cast <Chunk*> (i_p))
	{
		// ctor for the prg factory -- uses void * because of c interface

		connect ();
	}

public:

	ABCPtr
		(
		const ABCPtr& i_rhs
		) throw ()
  		:
		_chunk (i_rhs._chunk)
	{
		connect ();
	}

	~ABCPtr
		(
		) // throw whatever the dtor of AB (or derived classes) throws ...
	{
		// free of object is here - if needed
		// NO virtual ... we do not need a vtbl ...

		release ();
	}

	ABCPtr&
	operator=
  		(
		const ABCPtr& i_rhs
		) // throw whatever the dtor of object throws ...
	{
		// IMPORTANT : compare the values of chunk pointers,
      // not the addresses of the ptr objects...

		if (_chunk != i_rhs._chunk)
		{
			// first, free old object

			release ();

			_chunk = i_rhs._chunk;

			connect ();
		}

		return *this;
	}


private:

	void
	release
  		(
		)
	{
		// free object. decrement ref. counter, test 
		// state and destruct if needed

		disconnect ();

		if (get_count () == 0)
		{
			if (get_state () == Chunk::good)
			{
				destroy ();
			}
			else
			{
				set_state_deleted ();
				
				free_manager ();
			}
		}
	}

public:

	// access functions

	AB*
	get_dump_object_pointer // long name is a psycho ... ;-)
		(
		) const throw ()
	{
		// instead of a conversion operator - you 
		//have to do this explicitly

		return reinterpret_cast <AB*> (_chunk);
	}

	AB*
	operator->
		(
		) const throw ()
	{
		return reinterpret_cast <AB*> (_chunk);
	}

	int
	get_count
  		(
		) const throw ()
	{
		Chunk* const mem = _chunk - 1;

		return mem->_s._count;
	}

	void
	set_state
		(
		int i_s
		) const throw ()
	{
		Chunk* const mem = _chunk - 1;

		mem->_s._state = char (i_s);
	}

	void
	set_state_free
		(
		) const throw ()
	{
		set_state (Chunk::free);
	}

	void
	set_state_good
		(
		) const throw ()
	{
		set_state (Chunk::good);
	}

	void
	set_state_deleted
		(
		) const throw ()
	{
		set_state (Chunk::deleted);
	}

	void
	set_state_bad
		(
		) const throw ()
	{
		set_state (Chunk::bad);
	}

	Chunk::Memory_state
	get_state
		(
		) const throw ()
	{
		Chunk* const mem = _chunk - 1;

		return Chunk::Memory_state (mem->_s._state);
	}

	bool
	operator ==
		(
		const ABCPtr<AB>& i_rhs
		) const throw ()
	{
   	// same objects if address is same

		return _chunk == i_rhs._chunk;
	}

	bool
	operator <
		(
		const ABCPtr<AB>& i_rhs
		) const throw ()
	{
		// stl container needs this for ordering 
		// key values -- see map and set
		// assume contignous memory model here

		return (reinterpret_cast <char*> (_chunk)
			- reinterpret_cast <char*> (i_rhs._chunk)) < 0;
	}

private:

	// forbidden -- use ABCPtr_d instead
	ABCPtr
		(
		);

	void
	destroy
		(
		)
	{
		// IMPORTANT: this needs a virtual destructor for AB if AB is a
		// base class - otherwise this will not work correctly

		set_state_deleted ();
		
		Chunk* const mem = _chunk - 1;

		ABCPtrManager *manager = mem->_s._manager;

		try
		{
			AB* obj = reinterpret_cast <AB*> (_chunk);

			// please, dont throw an exception here - PLEASE !!

			obj->~AB ();		// BC 5.02 problem : obj->~AB (); doesnt work
		}

		catch (...)
		{
			free_memory ();	// no leaks

			manager->disconnect ();

			throw;
		}

		free_memory ();

		manager->disconnect ();
	}

	void
	free_memory
		(
		) throw ()
	{
		// free the memory via manager

		Chunk* const mem = _chunk - 1;

		mem->_s._manager->deallocate (mem);
	}

	void
	free_manager
		(
		) throw ()
	{
		// free the manager via manager

		Chunk* const mem = _chunk - 1;

		mem->_s._manager->disconnect ();
	}

	void
	connect
		(
		) const throw ()
	{
		// register a new connection to the memory

		Chunk* const mem = _chunk - 1;

		++ (mem->_s._count);
	}

	void
	disconnect
		(
		) const throw ()
	{
		// lose a connection

		Chunk* const mem = _chunk - 1;

		-- (mem->_s._count);
	}

}; // end ABCPtr


//------------------------------------------------------------------------

template < class AB >
class ABCPtr_d
{
	// abstract base reference counted pointer -- interface seen only
	// needs help of factory classes for instantiation
	// knows only the interface class AB. the implementation is
	// not known at the interface side of the system

	// has default ctor because of brain damaged libs and comps...

	Chunk* _chunk;

protected:

	explicit ABCPtr_d
		(
		void* i_p
		) throw (std::runtime_error)
		:
		_chunk (reinterpret_cast <Chunk*> (i_p))
	{
		// ctor for the prg factory -- use void * because of c interface

		// well - we dont allow this kind of shit ...
		if (_chunk == 0)
		{
			throw std::runtime_error(
      		std::string("illegal initialization of ABCPtr_d : void* is NULL"));
		}

		connect ();
	}

public:

	ABCPtr_d
		(
		) throw ()
		:
		_chunk (0)
	{
		// brain damaged libs and comps need this kind of stuff ...
		// for example take code warrior 4 and std::list<> ...
	}

	ABCPtr_d
		(
		const ABCPtr_d& i_rhs
		) throw ()
  		:
		_chunk (i_rhs._chunk)
	{
		if (_chunk)
		{
			connect ();
		}
	}

	ABCPtr_d
		(
		const ABCPtr <AB>& i_rhs
		) throw ()
  		:
		_chunk (reinterpret_cast <Chunk*> (i_rhs.get_dump_object_pointer ()))
	{
		// this direction is OK : assign / convert a ABCPtr to a ABCPtr_d
		// WARNING: the other direction is NOT OK -- dont do it ...

		connect ();
	}

	~ABCPtr_d
		(
		) // throw whatever the dtor of AB (or derived classes) throws ...
	{
		// free of object is here - if needed
		// NO virtual ... we do not need a vtbl ...

		if (_chunk)
		{
			release ();
		}
	}

	ABCPtr_d&
	operator=
  		(
		const ABCPtr_d& i_rhs
		) // throw whatever the dtor of object throws ...
	{
		// IMPORTANT : compare the values of chunks, not 
		// the addresses of selfs ...

		if (_chunk != i_rhs._chunk)
		{
			// first, free old object
			if (_chunk)
			{
				release ();
			}

			// now, copy address
			_chunk = i_rhs._chunk;

			// and correct count
			if (_chunk)
			{
				connect ();
			}
		}

		return *this;
	}


private:

	void
	release
  		(
		) // throw what AB dtor can throw ...
	{
		// free object. decrement ref. counter, test state 
		// and destruct if needed

		disconnect ();

		if (get_count () == 0)
		{
			if (get_state () == Chunk::good)
			{
				destroy ();
			}
			else
			{
				free_manager ();
			}
		}
	}

public:

	// access functions

	AB*
	get_dump_object_pointer // long name is a psycho ... ;-)
		(
		) const throw ()
	{
		// instead of a conversion operator - you have 
		// to do this explicitly
		return reinterpret_cast <AB*> (_chunk);
	}

	AB*
	operator->
		(
		) const throw ()
	{
		return reinterpret_cast <AB*> (_chunk);
	}

	int
	get_count
  		(
		) const throw ()
	{
		if (_chunk == 0)
		{
			return -1;
		}

		Chunk* const mem = _chunk - 1;

		return mem->_s._count;
	}

	void
	set_state
		(
		int i_s
		) const throw ()
	{
		if (_chunk == 0)
		{
			return ;
		}

		Chunk* const mem = _chunk - 1;

		mem->_s._state = char (i_s);
	}

	void
	set_state_free
		(
		) const throw ()
	{
		set_state (Chunk::free);
	}

	void
	set_state_good
		(
		) const throw ()
	{
		set_state (Chunk::good);
	}

	void
	set_state_deleted
		(
		) const throw ()
	{
		set_state (Chunk::deleted);
	}

	void
	set_state_bad
		(
		) const throw ()
	{
		set_state (Chunk::bad);
	}

	Chunk::Memory_state
	get_state
		(
		) const throw ()
	{
		if (_chunk == 0)
		{
			return Chunk::null;
		}

		Chunk* const mem = _chunk - 1;

		return Chunk::Memory_state (mem->_s._state);
	}

	bool
	operator ==
		(
		const ABCPtr_d <AB>& i_rhs
		) const throw ()
	{
		return _chunk == i_rhs._chunk;
	}

	bool
	operator <
		(
		const ABCPtr_d <AB>& i_rhs
		) const throw ()
	{
		// well. this is needed for ordered containers with a abcptr_d as key ...
		// for example a set or map ...

		return (reinterpret_cast <char*> (_chunk)
      	- reinterpret_cast <char*> (i_rhs._chunk)) < 0;
	}

private:

	void
	destroy
		(
		)
	{
		// IMPORTANT: this needs a virtual destructor for AB if AB is a
		// base class - otherwise this will not work correctly ...

		Chunk* const mem = _chunk - 1;

		// we need this later after deletion of tile 0 - so save it now
		ABCPtrManager *manager = mem->_s._manager;

		try
		{
			AB* obj = reinterpret_cast <AB*> (_chunk);

			// please, dont throw an exception here - PLEASE !!

			obj->~AB ();		// BC 5.02 problem : obj->~AB (); doesnt work
								// use obj->AB::~AB() instead...
		}

		catch (...)
		{
			// but if you do, we are ready for it ...
			free_memory ();	// no leaks

			manager->disconnect ();

			throw;
		}

		free_memory ();

		manager->disconnect ();
	}

	void
	free_memory
		(
		) throw ()
	{
		// free the memory via manager

		Chunk* const mem = _chunk - 1;

		mem->_s._manager->deallocate (mem);
	}

	void
	free_manager
		(
		) throw ()
	{
		// free the manager via manager

		Chunk* const mem = _chunk - 1;

		mem->_s._manager->disconnect ();
	}

	void
	connect
		(
		) const throw ()
	{
		// register a new connection to the memory

		Chunk* const mem = _chunk - 1;

		++ (mem->_s._count);
	}

	void
	disconnect
		(
		) const throw ()
	{
		// lose a connection

		Chunk* const mem = _chunk - 1;

		-- (mem->_s._count);
	}

}; // end ABCPtr_d


} // end Utility

} // end AR

//----------------------------------------------------------------------

// poor man solution for inplace construction of a abcptr
// and usage of any ctor of the derived class
// example: ABCPTR(X, ptr, Y, (parm1,parm2,parm3));
// where Y is a derived class from class X
// and has a ctor of the form Y (T1 parm1, T2 parm2, T3 parm3)
// because you have to know the ctor you have to include
// interface and implementation together
// you have also to include abcptrinit and abcptrdllinit
// for the helper classes - its really a poor solution...

#ifndef ABCPTR

//#include <ar/utility/abcptrinit.h>
//#include <ar/utility/abcptrdllinit.h>

// use a manager
#define ABCPTR_MANAGER(__classX,__namePtr,__classY,__params,__manager) \
	AR::Utility::ABCPtr < __classX > __namePtr ( \
   	AR::Utility::ABCPtrInit < __classX > \
      	( \
         	new (AR::Utility::ABCPtrDLLInit ( \
		    		((sizeof(__classY) + sizeof(AR::Utility::Chunk) - 1) \
			/ sizeof (AR::Utility::Chunk) + 1) \
			* sizeof(AR::Utility::Chunk), \
				__manager).get_adr ()) __classY __params \
         ) \
      )

// use the default manager
#define ABCPTR(__classX,__namePtr,__classY,__params) \
	ABCPTR_MANAGER(__classX,__namePtr,__classY,__params,  \
   		AR::Utility::g_default_manager_p)

#endif

#endif

// end of file

