LISTING 5

#ifndef INCLUDED_AR_UTILITY_ABCPTRDLLINIT_H
#define INCLUDED_AR_UTILITY_ABCPTRDLLINIT_H

// copyright 1998 by anton rommerskirchen
// this code can be used and copied for free without changes
// no warrenty is given by the author
// enhancements to arommerskirchen@csi.com
// $Id: abcptrdllinit.h,v 1.1 2001/10/03 12:45:19 ar Exp $

#ifndef INCLUDED_STDEXCEPT
#	include <stdexcept>
#	undef INCLUDED_STDEXCEPT
#	define INCLUDED_STDEXCEPT
#endif

#ifndef INCLUDED_AR_UTILITY_ABCPTRMANAGER_H
#	include <ar/utility/abcptrmanager.h>
#	undef INCLUDED_AR_UTILITY_ABCPTRMANAGER_H
#	define INCLUDED_AR_UTILITY_ABCPTRMANAGER_H
#endif


#ifndef INCLUDED_AR_UTILITY_CHUNK_H
#	include <ar/utility/chunk.h>
#	undef INCLUDED_AR_UTILITY_CHUNK_H
#	define INCLUDED_AR_UTILITY_CHUNK_H
#endif

namespace AR {

namespace Utility {

//------------------------------------------------------------------------------

class ABCPtrImpl
{
	// base class for management of handles
	// there are no virtual functions ...
	
protected:

	Chunk* _chunk;

protected:
	// ctors

	explicit ABCPtrImpl
  		(
		Chunk* i_chunk,
		ABCPtrManager*	i_manager
		) throw ()
		:
		_chunk (i_chunk + 1)
	{
		// initialize with an closed array of chunks
		// but do not use the address of the array,
		// use instead the address of the second chunk
		// this is the major trick

		i_chunk->_s._manager = i_manager;

		i_chunk->_s._count = 0;

		i_chunk->_s._state = Chunk::free;

      // length is set by allocator -- we dont know our length ...
	}


	explicit ABCPtrImpl
  		(
		void* i_p
		) throw ()
		:
		_chunk (reinterpret_cast <Chunk*> (i_p))
	{
		// use this inside the factorys
	}

	ABCPtrImpl
		(
		const ABCPtrImpl& i_rhs
		) throw ()
		:
		_chunk (i_rhs._chunk)
	{
		// copy is simple. ref counting is done at a upper level
	}

	~ABCPtrImpl
		(
		) throw ()
	{
		// NO virtual ....
		// we do not need a vtbl for any of this classes ....
		// free of memory is NOT here ...
		// ... this is done at a higher level for placement requirements
	}


	void
	connect
		(
		) const throw ()
	{
		// register a new connection to the memory

		Chunk* const mem = _chunk - 1;

		++ (mem->_s._count);
	}

	void
	disconnect
		(
		) const throw ()
	{
		// lose a connection

		Chunk* const mem = _chunk - 1;

		-- (mem->_s._count);
	}

	void
	free_memory
		(
		) const throw ()
	{
		// free the memory via manager

		Chunk* const mem = _chunk - 1;

		mem->_s._manager->deallocate (mem);
	}

	void
	free_manager
		(
		) const throw ()
	{
		// free the manager via manager

		Chunk* const mem = _chunk - 1;

		mem->_s._manager->disconnect ();
	}

	void
	set_state
		(
		int i_s
		) const throw ()
	{
		Chunk* const mem = _chunk - 1;

		mem->_s._state = char (i_s);
	}

	void
	set_state_free
		(
		) const throw ()
	{
		set_state (Chunk::free);
	}

	void
	set_state_good
		(
		) const throw ()
	{
		set_state (Chunk::good);
	}

	void
	set_state_deleted
		(
		) const throw ()
	{
		set_state (Chunk::deleted);
	}

	void
	set_state_bad
		(
		) const throw ()
	{
		set_state (Chunk::bad);
	}

	Chunk::Memory_state
	get_state
		(
		) const throw ()
	{
		Chunk* const mem = _chunk - 1;

		return Chunk::Memory_state (mem->_s._state);
	}

	int
	get_count
  		(
		) const throw ()
	{
		Chunk* const mem = _chunk - 1;

		return mem->_s._count;
	}

}; // end ABCPtrImpl


//--------------------------------------------------------------------------------

class ABCPtrDLLInit
	:
	private ABCPtrImpl
{
	// class for constuction of a reference counted
	// pointer in the dll factory. has to know nothing
	// about the interface and implementation classes

public:

	// ctors

	explicit ABCPtrDLLInit
		(
		int i_size,
		ABCPtrManager* i_manager
		) // throw whatever allocate can throw
		:
		ABCPtrImpl (i_manager->allocate (i_size), i_manager)
	{
		// reserve the memory via allocation function and
		// initialize management info
		// aligning must be correct...
		// free the memory by deallocate callback object function
	}

	~ABCPtrDLLInit
		(
		) throw ()
	{
		// NO virtual - we do not need the vtbl ....
	}

	void*
	get_adr
		(
		) const throw ()
	{
		// deliver the memory address -- for placement new in the dll factory

		return _chunk;
	}

	// helpers

	void
	free_memory
		(
		) const throw ()
	{
		ABCPtrImpl::free_memory ();
	}

private:

	// not allowed
	ABCPtrDLLInit ();

	ABCPtrDLLInit (const ABCPtrDLLInit &);

}; // end ABCPtrDLLInit

} // end Utility

} // end AR

#endif

// end of file

