// copyright 1998 by anton rommerskirchen
// this code can be used and copied for free without changes
// no warrenty is given by the author
// enhancements to arommerskirchen@csi.com
// $Id: mymanager.cpp,v 1.1 2001/10/03 12:45:19 ar Exp $

//
// implementation of a manager
// use new and delete
// use loadlibrary and freelibrary
//

#ifndef INCLUDED_MYMANAGER_H
#	include "mymanager.h"
#	undef INCLUDED_MYMANAGER_H
#	define INCLUDED_MYMANAGER_H
#endif

#ifndef INCLUDED_STDEXCEPT
#	include <stdexcept>
#	undef INCLUDED_STDEXCEPT
#	define INCLUDED_STDEXCEPT
#endif

#ifndef INCLUDED_STRING
#	include <string>
#	undef INCLUDED_STRING
#	define INCLUDED_STRING
#endif

#if __WINDOWS__
// we need this for the dll loadings ...
#ifndef INCLUDED_WINDOWS_H
#	include <windows.h>
#	undef INCLUDED_WINDOWS_H
#	define INCLUDED_WINDOWS_H
#endif
#endif

// we are on a linux box
#if __GNUG__

#include <dlfcn.h>
#endif

//----------------------------------------------------------------------------

void
Mymanager::connect
	(
	) throw (std::overflow_error)
{
	// count the actual connected objects, not the refs ...

	++_refs;

	// ups - a hard coded limit ...
	if (_refs >= 32767)
	{
		throw std::overflow_error ("ups, to many objects for this dll ...");
	}
}

//----------------------------------------------------------------------------

void
Mymanager::disconnect
	(
	) throw ()
{
	// free the dll reference. win32 version

	// new count of actual connected objects for this dll
	decrement ();

	if (_refs <= 0)
	{
		// we do not need it any more ...
		if (_dll_handle)
		{
#if __windows__
			BOOL free_erg = ::FreeLibrary (_dll_handle);
#endif
#if __GNUG__
			bool free_erg = ::dlclose (_dll_handle);
#endif
		}

		_dll_handle = 0;

		_factory = 0;
	}
}

//----------------------------------------------------------------------------

void
Mymanager::decrement
	(
	) throw ()
{
	// do not free the dll

	// new count of actual connected objects for this dll
	--_refs;
}

//------------------------------------------------------------------

Mymanager::Pfunc
Mymanager::get_factory
	(
	)
{
	// load the lib -- if needed. win32 version

	if (_dll_handle == 0)
	{
#if __windows__
		if ((_dll_handle = ::LoadLibrary (_name_of_dll.c_str())) == 0)
			throw std::runtime_error ("TODO: DLL x.dll NOT FOUND ");
#endif
#if __GNUG__
		if ((_dll_handle = ::dlopen (_name_of_dll.c_str(), RTLD_NOW)) == 0)
		  {
		    char* errtext = dlerror();
		    std::cerr << errtext << std::endl;
			throw std::runtime_error ("TODO: DLL x.dll NOT FOUND ");
		  }
#endif

		_factory = 0;
	}

	if (_factory == 0)
	{
#if __windows__
		if ((_factory = reinterpret_cast <Pfunc> (::GetProcAddress (_dll_handle,
				_name_of_factory.c_str ()))) == 0)
		{
			::FreeLibrary (_dll_handle);

			throw std::runtime_error ("TODO: FUNCTION makeAX_InDll not found");
		}
#endif
#if __GNUG__
		if ((_factory = (Pfunc) (::dlsym (_dll_handle,
				_name_of_factory.c_str ()))) == 0)
		{
			::dlclose (_dll_handle);

			throw std::runtime_error ("TODO: FUNCTION makeAX_InDll not found");
		}
#endif
	}

	return _factory;
}

//------------------------------------------------------------------

Mymanager::~Mymanager
	(
	)
{
	// if needed, free the dll

	disconnect ();
}

//------------------------------------------------------------------

// end of file


