// copyright 1998 by anton rommerskirchen
// this code can be used and copied for free without changes
// no warrenty is given by the author
// enhancements to arommerskirchen@csi.com
// $Id: withmacros.cpp,v 1.1 2001/10/03 12:45:19 ar Exp $

//
// test for abstract base reference counting smart pointers
// interface and implementation are physical separated in
// prg and dll. dll ist loaded on demand at runtime
// and freed as early as possible
//


// to see is to believe ...
#ifndef INCLUDED_IOSTREAM
#	include <iostream>
#	undef INCLUDED_IOSTREAM
#	define INCLUDED_IOSTREAM
#endif

// exception classes are here
#ifndef INCLUDED_STDEXCEPT
#	include <stdexcept>
#	undef INCLUDED_STDEXCEPT
#	define INCLUDED_STDEXCEPT
#endif

// stl list -- try it
#ifndef INCLUDED_LIST
#	include <list>
#	undef INCLUDED_LIST
#	define INCLUDED_LIST
#endif

#ifndef INCLUDED_MAP
#	include <map>
#	undef INCLUDED_MAP
#	define INCLUDED_MAP
#endif

#ifndef INCLUDED_SET
#	include <set>
#	undef INCLUDED_SET
#	define INCLUDED_SET
#endif

// the user needs only this
#ifndef INCLUDED_AR_UTILITY_ABCPTR_H
#	include <ar/utility/abcptr.h>
#	undef INCLUDED_AR_UTILITY_ABCPTR_H
#	define INCLUDED_AR_UTILITY_ABCPTR_H
#endif

#ifndef INCLUDED_AR_UTILITY_ABCPTRDLLINIT_H
#	include <ar/utility/abcptrdllinit.h>
#	undef INCLUDED_AR_UTILITY_ABCPTRDLLINIT_H
#	define INCLUDED_AR_UTILITY_ABCPTRDLLINIT_H
#endif

#ifndef INCLUDED_AR_UTILITY_ABCPTRINIT_H
#	include <ar/utility/abcptrinit.h>
#	undef INCLUDED_AR_UTILITY_ABCPTRINIT_H
#	define INCLUDED_AR_UTILITY_ABCPTRINIT_H
#endif

// the interface class
#ifndef INCLUDED_X_H
#	include "x.h"
#	undef INCLUDED_X_H
#	define INCLUDED_X_H
#endif

#ifndef INCLUDED_Y_H
#	include "y.h"
#	undef INCLUDED_Y_H
#	define INCLUDED_Y_H
#endif


//------------------------------------------------------------------

// try a stl container with it

typedef std::list <AR::Utility::ABCPtr_d <X> > list_of_X_ptr;

//------------------------------------------------------------------

// try to be a subobject of another class

class Z
{
	AR::Utility::ABCPtr_d <X> _x;

public:

	Z
		(
		const AR::Utility::ABCPtr_d <X>& i_x
		)
		:
		_x (i_x)
	{
	}

	~Z
		(
		)
	{
		std::cout << "dtor Z" << std::endl;
	}

	void
	doit
		(
		) const
	{
		_x->doit ();
	}
}; // end Z

//------------------------------------------------------------------

namespace {

//------------------------------------------------------------------

void
hugotest_by_value
	(
	AR::Utility::ABCPtr <X> i_p
	)
{
	// use a abc ptr by value

	std::cout << "by value : " << i_p.get_count () << std::endl;

	i_p->doit ();
}

//------------------------------------------------------------------

void
hugotest_by_ref
	(
	const AR::Utility::ABCPtr <X>& i_p
	)
{

	// use a abc ptr by reference

	std::cout << "by ref : " << i_p.get_count () << std::endl;

	i_p->doit ();
}

//------------------------------------------------------------------

} // end anonym

//------------------------------------------------------------------

int
main
	(
	int argc,
	char* argv[]
	)
{
	try
	{
		ABCPTR (X, p42, Y, (4812));
      		
		p42->doit ();
      
		// dump pointer
		X* hansi = p42.get_dump_object_pointer ();

		const X* hansi_c = p42.get_dump_object_pointer ();

		X* const hansi_c2 = p42.get_dump_object_pointer ();

		hansi->doit ();

		// now, its time to do something more interesting
		std::cout << p42.get_count () << std::endl;

		hugotest_by_value (p42);

		std::cout << p42.get_count () << std::endl;

		hugotest_by_ref (p42);

		std::cout << p42.get_count () << std::endl;

		p42->doit ();

		// well. now its time for a container
		{

			list_of_X_ptr l;

			l.push_back (p42);

			l.push_back (p42);

			std::cout << "list " << p42.get_count () << std::endl;

			for (list_of_X_ptr::iterator it = l.begin ();
					it != l.end ();
					++it)
			{
				(*it)->doit (); // cruel, but works

			}
		}

		std::cout << "after list dtor " << p42.get_count () << std::endl;

		typedef AR::Utility::ABCPtr_d<X> Ptr;

		std::map <Ptr , int, std::less <Ptr> > m;

		std::set <Ptr , std::less <Ptr> > s;

		m[p42] = 1;

		s.insert(p42);
      
		// test it inside another object
		{
			Z z(p42);

			std::cout << "inside Z " << p42.get_count () << std::endl;

			z.doit ();
		}

		std::cout << "after Z dtor " << p42.get_count () << std::endl;
	} // dtor of X 42 ...

	catch (const char* i_text)
	{
		std::cout << i_text << std::endl;
	}


	catch (std::exception& e)
	{
		std::cout << e.what () << std::endl;
	}

	catch (double d)
	{
		std::cout << "exception " << d << std::endl;
	}

	catch (...)
	{
		std::cout << "exception ..." << std::endl;

	}

	return 0;
}

// end of file

