// $Id: withdlls.cpp,v 1.2 2001/10/10 06:34:27 ar Exp $

/* Copyright (C) 2001 ATRSoft GmbH 
   This file is part of the ATRSoft GmbH utilitys.

   The ATRSoft GmbH utilitys is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or any later version.

   The ATRSoft GmbH utilitys is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with the ATRSoft GmbH utilitys; see the file LICENSE.  If not,
   write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

// http://www.atrsoft.de
// mailto: info@atrsoft.de

// visual c++ mfc add on. we dont need this if no afx app....
// for mfc apps you have to remove the double slashs 
// #include <stdafx.h>

//
// test for abstract base reference counting smart pointers
// interface and implementation are physical separated in
// prg and dll. dll ist loaded on demand at runtime
// and freed as early as possible
//


// to see is to believe ...
#ifndef INCLUDED_IOSTREAM
#	include <iostream>
#	undef INCLUDED_IOSTREAM
#	define INCLUDED_IOSTREAM
#endif

// exception classes are here
#ifndef INCLUDED_STDEXCEPT
#	include <stdexcept>
#	undef INCLUDED_STDEXCEPT
#	define INCLUDED_STDEXCEPT
#endif

// stl list -- try it
#ifndef INCLUDED_LIST
#	include <list>
#	undef INCLUDED_LIST
#	define INCLUDED_LIST
#endif

#ifndef INCLUDED_MAP
#	include <map>
#	undef INCLUDED_MAP
#	define INCLUDED_MAP
#endif

#ifndef INCLUDED_SET
#	include <set>
#	undef INCLUDED_SET
#	define INCLUDED_SET
#endif

// the user needs only this
#ifndef INCLUDED_ATRSOFT_UTILITY_INCLUDE_ATRP_H
#	include <atrsoft/utility/include/atrp.h>
#	undef INCLUDED_ATRSOFT_UTILITY_INCLUDE_ATRP_H
#	define INCLUDED_ATRSOFT_UTILITY_INCLUDE_ATRP_H
#endif

// the object factory is there 
// could be an object file, a lib ...
#ifndef INCLUDED_ATRPFACT_H
#	include "atrpfact.h"
#	undef INCLUDED_ATRPFACT_H
#	define INCLUDED_ATRPFACT_H
#endif

// the manager for the so and the object memory
#ifndef INCLUDED_MYMANAGER_H
#	include "mymanager.h"
#	undef INCLUDED_MYMANAGER_H
#	define INCLUDED_MYMANAGER_H
#endif

// the interface class
#ifndef INCLUDED_X_H
#	include "x.h"
#	undef INCLUDED_X_H
#	define INCLUDED_X_H
#endif

// not needed : y 
// we dont have to know anything about it - not even that it exists
// only x is needed. physical seperation of the prg and y is made
// by a shared object loader with a factory 

//------------------------------------------------------------------

// try a stl container with it

typedef std::list <ATRSoft::Utility::Atrp_d <X> > list_of_X_ptr;

#if ATRP_ON_LINUX
const char* g_soname = "x.so";
const char* g_factname = "makeAX_InImpl";
#endif

#if ATRP_ON_WINDOWS
const char* g_soname = "xabcptr.dll";
const char* g_factname = "makeAX_InImpl";
#endif

// MSVC & Code Warrior name. Use map file to see YOUR name ...

//------------------------------------------------------------------

// try to be a subobject of another class

class Z
{
    ATRSoft::Utility::Atrp_d <X> _x;

public:

    Z
        (
	const ATRSoft::Utility::Atrp_d <X>& i_x
	)
	:
	_x (i_x)
    {
    }

    ~Z
        (
	)
    {
	std::cout << "dtor Z" << std::endl;
    }

    void
    doit
        (
	) const
    {
	_x->doit ();
    }
}; // end Z

//------------------------------------------------------------------

namespace {

//------------------------------------------------------------------

void
hugotest_by_value
    (
    ATRSoft::Utility::Atrp_d <X> i_p
    )
{
    // use a abc ptr by value

    std::cout << "by value : " << i_p.get_count () << std::endl;

    i_p->doit ();
}

//------------------------------------------------------------------

void
hugotest_by_ref
    (
    const ATRSoft::Utility::Atrp_d <X>& i_p
    )
{
    // use a abc ptr by reference

    std::cout << "by ref : " << i_p.get_count () << std::endl;

    i_p->doit ();
}

//------------------------------------------------------------------

} // end anonym

//------------------------------------------------------------------

int
main
    (
    int argc,
    char* argv[]
    )
{
    // create a manager which uses the x.so and within the
    // makeAX_InImpl factory
    Mymanager m (g_soname, g_factname);


    try
    {
	// create a X reference, value ist 42
	ATRSoft::Utility::Atrp_d <X> p (makeAX (42, &m));

	// make a copy
	ATRSoft::Utility::Atrp_d <X> p2 = p;

	// another X object, value ist 99, same manager
	ATRSoft::Utility::Atrp_d <X> p3 (makeAX (99, &m));

	ATRSoft::Utility::Atrp_d <X> p4; // null test ...

	ATRSoft::Utility::Atrp_d <X> p5; // null test ...

	// assign - set X 99 free ...
	p3 = p2;

	// assign self
	p2 = p3;


	p4 = p2;


	p4 = p5;

	// use X ...
	p->doit ();

	p2->doit ();

	p3->doit ();

	// dump pointer
	X* hansi = p2.get_dump_object_pointer ();

	const X* hansi_c = p2.get_dump_object_pointer ();

	X* const hansi_c2 = p2.get_dump_object_pointer ();

	hansi->doit ();

	// now, its time to do something more interesting
	std::cout << p.get_count () << std::endl;

	hugotest_by_value (p);

	std::cout << p.get_count () << std::endl;

	hugotest_by_ref (p);

	std::cout << p.get_count () << std::endl;

	p->doit ();

	// well. now its time for a container
	{

	    list_of_X_ptr l;

	    l.push_back (p);

	    l.push_back (p2);

	    std::cout << "list " << p.get_count () << std::endl;

	    for (list_of_X_ptr::iterator it = l.begin ();
		 it != l.end ();
		 ++it)
	    {
		(*it)->doit (); // cruel, but works

	    }
	}

	std::cout << "after list dtor " << p.get_count () << std::endl;

	typedef ATRSoft::Utility::Atrp_d<X> Ptr;

	std::map <Ptr , int, std::less <Ptr> > m;

	std::set <Ptr , std::less <Ptr> > s;

	m[p] = 1;

	s.insert(p);
      
	// test it inside another object
	{
	    Z z(p);

	    std::cout << "inside Z " << p.get_count () << std::endl;

	    z.doit ();
	}

	std::cout << "after Z dtor " << p.get_count () << std::endl;
    } // dtor of X 42 ...

    catch (const char* i_text)
    {
	std::cout << i_text << std::endl;
    }


    catch (std::exception& e)
    {
	std::cout << e.what () << std::endl;
    }

    catch (double d)
    {
	std::cout << "exception " << d << std::endl;
    }

    catch (...)
    {
	std::cout << "exception ..." << std::endl;
	
    }

    return 0;
}

// end of file
