#ifndef INCLUDED_ATRSOFT_UTILITY_INCLUDE_ATRPIMPL_H
#define INCLUDED_ATRSOFT_UTILITY_INCLUDE_ATRPIMPL_H


// $Id: atrpimpl.h,v 1.6 2001/10/04 12:27:36 ar Exp $
/* Copyright (C) 2001 ATRSoft GmbH 
   This file is part of the ATRSoft GmbH utilitys.

   The ATRSoft GmbH utilitys is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or any later version.

   The ATRSoft GmbH utilitys is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with the ATRSoft GmbH utilitys; see the file LICENSE.  If not,
   write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

// http://www.atrsoft.de
// mailto: info@atrsoft.de


#ifndef INCLUDED_ATRSOFT_UTILITY_INCLUDE_ATRP_H
#	include <atrsoft/utility/include/atrp.h>
#	undef INCLUDED_ATRSOFT_UTILITY_INCLUDE_ATRP_H
#	define INCLUDED_ATRSOFT_UTILITY_INCLUDE_ATRP_H
#endif

namespace ATRSoft {

namespace Utility {

//------------------------------------------------------------------------------

class AtrpImpl
{
    // base class for management of handles
    // there are no virtual functions ...
	
protected:

    Chunk* chunk_;

protected:
    // ctors

    explicit AtrpImpl
        (
	Chunk* i_chunk,
	AtrpManager*	i_manager
	) throw ()
        :
	chunk_ (i_chunk + 1)
    {
	// initialize with an closed array of chunks
	// but do not use the address of the array,
	// use instead the address of the second chunk
	// this is the major trick

	i_chunk->s_.manager_ = i_manager;

	i_chunk->s_.count_ = 0;
    
	i_chunk->s_.state_ = Chunk::free;

	// length is set by allocator -- we dont know our length ...
    }


    explicit AtrpImpl
        (
	void* i_p
	) throw ()
	:
	chunk_ (reinterpret_cast <Chunk*> (i_p))
    {
	// use this inside the factorys
    }

    AtrpImpl
        (
	const AtrpImpl& i_rhs
	) throw ()
	:
	chunk_ (i_rhs.chunk_)
    {
	// copy is simple. ref counting is done at a upper level
    }

    ~AtrpImpl
        (
	) throw ()
    {
	// NO virtual ....
	// we do not need a vtbl for any of this classes ....
	// free of memory is NOT here ...
	// ... this is done at a higher level for placement requirements
    }


    void
    connect
        (
	) const throw ()
    {
	// register a new connection to the memory

	Chunk* const mem = chunk_ - 1;

	++ (mem->s_.count_);
    }
    
    void
    disconnect
        (
	) const throw ()
    {
	// lose a connection

	Chunk* const mem = chunk_ - 1;

	-- (mem->s_.count_);
    }

    void
    free_memory
        (
	) const throw ()
    {
	// free the memory via manager

	Chunk* const mem = chunk_ - 1;

	mem->s_.manager_->deallocate (mem);
    }

    void
    free_manager
        (
	) const throw ()
    {
	// free the manager via manager

	Chunk* const mem = chunk_ - 1;

	mem->s_.manager_->disconnect ();
    }

    void
    set_state
        (
	int i_s
	) const throw ()
    {
	Chunk* const mem = chunk_ - 1;
	
	mem->s_.state_ = char (i_s);
    }
    
    void
    set_state_free
        (
	) const throw ()
    {
	set_state (Chunk::free);
    }

    void
    set_state_good
        (
	) const throw ()
    {
	set_state (Chunk::good);
    }

    void
    set_state_deleted
        (
	) const throw ()
    {
	set_state (Chunk::deleted);
    }

    void
    set_state_bad
        (
	) const throw ()
    {
	set_state (Chunk::bad);
    }

    Chunk::Memory_state
    get_state
        (
	) const throw ()
    {
	Chunk* const mem = chunk_ - 1;
	
	return Chunk::Memory_state (mem->s_.state_);
    }

    int
    get_count
        (
	) const throw ()
    {
	Chunk* const mem = chunk_ - 1;
	    
	return mem->s_.count_;
    }

}; // end AtrpImpl


//--------------------------------------------------------------------------------

class AtrpImplInit
    :
    private AtrpImpl
{
    // class for constuction of a reference counted
    // pointer in the dll factory. has to know nothing
    // about the interface and implementation classes

public:

    // ctors

    explicit AtrpImplInit
        (
	int i_size,
	AtrpManager* i_manager
	) // throw whatever allocate can throw
	:
        AtrpImpl (i_manager->allocate (i_size), i_manager)
    {
	// reserve the memory via allocation function and
	// initialize management info
	// aligning must be correct...
	// free the memory by deallocate callback object function
    }

    ~AtrpImplInit
        (
	) throw ()
    {
	// NO virtual - we do not need the vtbl ....
    }
 
    void*
    get_adr
        (
	) const throw ()
    {
	// deliver the memory address -- for placement new in the dll factory

	return chunk_;
    }
    
    // helpers

    void
    free_memory
        (
	) const throw ()
    {
	AtrpImpl::free_memory ();
    }

private:

    // not allowed
    AtrpImplInit ();

    AtrpImplInit (const AtrpImplInit &);

}; // end AtrpImplInit

} // end Utility

} // end Atrsoft

#endif

// end of file
