// $Id: mymanager.cpp,v 1.6 2001/10/10 06:34:27 ar Exp $

/* Copyright (C) 2001 ATRSoft GmbH 
   This file is part of the ATRSoft GmbH utilitys.

   The ATRSoft GmbH utilitys is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or any later version.

   The ATRSoft GmbH utilitys is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with the ATRSoft GmbH utilitys; see the file LICENSE.  If not,
   write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

// http://www.atrsoft.de
// mailto: info@atrsoft.de

// visual c++ mfc add on. we dont need this if no afx app....
// for mfc apps you have to remove the double slashs 
// #include <stdafx.h>

//
// implementation of a manager
// use new and delete
// use loadlibrary and freelibrary
//

#ifndef INCLUDED_MYMANAGER_H
#	include "mymanager.h"
#	undef INCLUDED_MYMANAGER_H
#	define INCLUDED_MYMANAGER_H
#endif

#ifndef INCLUDED_STDEXCEPT
#	include <stdexcept>
#	undef INCLUDED_STDEXCEPT
#	define INCLUDED_STDEXCEPT
#endif

#ifndef INCLUDED_STRING
#	include <string>
#	undef INCLUDED_STRING
#	define INCLUDED_STRING
#endif

#ifndef INCLUDED_IOSTREAM
#       include <iostream>
#       undef INCLUDED_IOSTREAM
#       define INCLUDED_IOSTREAM
#endif

// dynamisch laden unter linux

#define ATRP_ON_LINUX 1
#define ATRP_ON_WINDOWS 0

#if ATRP_ON_LINUX
#ifndef INCLUDED_DLFCN_h
#       include <dlfcn.h>
#       undef INCLUDED_DLFCN_h
#       define INCLUDED_DLFCN_h
#endif
#endif

#if ATRP_ON_WINDOWS
// we need this for the dll loadings ...
#ifndef INCLUDED_WINDOWS_H
#	include <windows.h>
#	undef INCLUDED_WINDOWS_H
#	define INCLUDED_WINDOWS_H
#endif
#endif

//----------------------------------------------------------------------------

void
Mymanager::connect
    (
    ) throw ()
{
    // count the actual connected objects, not the refs ...

    ++refs_;
}

//----------------------------------------------------------------------------

void
Mymanager::disconnect
    (
    ) throw ()
{
    try
    {
	// free the so reference. 

	// new count of actual connected objects for this dll
	decrement ();

	if (refs_ <= 0)
	{
	    // we do not need it any more ...
	    if (handle_)
	    {
#if ATRP_ON_LINUX
		bool free_erg = ::dlclose (handle_) == 0;
#endif

#if ATRP_ON_WINDOWS
		bool free_erg = ::FreeLibrary (handle_) == 0;
#endif
		if (free_erg)
		{
		    std::string err = "dlclose for ";
		    err += name_of_so_;

		    std::cerr << err << std::endl;
		}
	    }

	    handle_ = 0;

	    factory_ = 0;
	}
    }
    catch (...)
    {
    }
}
//----------------------------------------------------------------------------

void
Mymanager::decrement
    (
    ) throw ()
{
    // do not free the dll

    // new count of actual connected objects for this dll
    --refs_;
}

//------------------------------------------------------------------

// ok
// das kann man beliebig komplizieren. z.b. ist eine registry aller geladenen
// factories denkbar (und sinnvoll !) mit manipulation der 
// gelieferten factory zur laufzeit. pflege ist so realisierbar
// in dieser version nur aufmachen des shared object und dann holen der
// factory funktion. mehr nicht.

Mymanager::Pfunc
Mymanager::get_factory
    (
    ) throw (std::runtime_error)
{
    // load the lib -- if needed. 

    if (handle_ == 0)
    {
#if ATRP_ON_LINUX
	handle_ = ::dlopen (name_of_so_.c_str(), RTLD_NOW);
#endif

#if ATRP_ON_WINDOWS
	handle_ = ::LoadLibrary (name_of_so_.c_str());
#endif
	if (handle_ == 0)
	{
#if ATRP_ON_LINUX
	    char* errtext = ::dlerror();
#endif

#if ATRP_ON_WINDOWS
	    char* errtext = "Symbol not found : ";
#endif

	    std::string err = "dlopen failed for";
	    err += name_of_so_;
	    err += " , ";
	    err += errtext;
	    throw std::runtime_error (err);
	}

	factory_ = 0;
    }

    // find the factory -- if needed.
    if (factory_ == 0)
    {
#if ATRP_ON_LINUX
	factory_ = (Pfunc) (::dlsym (handle_, name_of_factory_.c_str ()));
#endif
	
#if ATRP_ON_WINDOWS
	factory_ = (Pfunc)(::GetProcAddress (handle_, name_of_factory_.c_str ()));
#endif
	if (factory_  == 0)
	{
#if ATRP_ON_LINUX
	    ::dlclose (handle_);
#endif

#if ATRP_ON_WINDOWS
	    ::FreeLibrary (handle_);
#endif

	    std::string err = "dlsym failed for ";
	    err += name_of_factory_;
	    err += " in " ;
	    err += name_of_so_;

	    throw std::runtime_error (err);
	}
    }

    return factory_;
}

//------------------------------------------------------------------

Mymanager::~Mymanager
    (
    ) throw ()
{
    try
    {
	// if needed, free the dll

	disconnect ();
    }
    catch (...)
    {
    }
}
//------------------------------------------------------------------

// end of file
